@@ -1,29 +1,26 @@
+from __future__ import annotations
+
 import random
 import string
-from typing import Optional
-
-from colorama import Style, init
+from typing import TYPE_CHECKING, Any, Iterable, List, Tuple
 
 from word_search_generator import config
-from word_search_generator.types import Key, Puzzle
+from word_search_generator.types import (
+    Direction,
+    DirectionSet,
+    Key,
+    Puzzle,
+    Word,
+    Wordlist,
+)
 
-init()
+if TYPE_CHECKING:  # pragma: no cover
+    from word_search_generator import WordSearch
 
 
-def cleanup_input(words: str) -> set[str]:
+def cleanup_input(words: str, secret: bool = False) -> Wordlist:
     """Cleanup provided input string. Removing spaces
-    one-letter words, and words with punctuation.
-
-    Args:
-        words (str): String of words separated by commas, spaces, or new lines.
-
-    Raises:
-        TypeError: A string was not provided.
-        ValueError: No proper words were provided.
-
-    Returns:
-        set[str]: Words to be placed in the puzzle.
-    """
+    one-letter words, and words with punctuation."""
     if not isinstance(words, str):
         raise TypeError(
             "Words must be a string separated by spaces, commas, or new lines"
@@ -33,11 +30,16 @@ def cleanup_input(words: str) -> set[str]:
     # remove excess spaces and commas
     word_list = ",".join(words.split(" ")).split(",")
     # iterate through all words and pick first set that match criteria
-    word_set: set[str] = set()
+    word_set: Wordlist = set()
     while word_list and len(word_set) <= config.max_puzzle_words:
         word = word_list.pop(0)
-        if len(word) > 1 and not contains_punctuation(word):
-            word_set.add(word.upper())
+        if (
+            len(word) > 1
+            and not contains_punctuation(word)
+            and not is_palindrome(word)
+            and not word_contains_word(word_set, word.upper())
+        ):
+            word_set.add(Word(word, secret=secret))
     # if no words were left raise exception
     if not word_set:
         raise ValueError("Use words longer than one-character and without punctuation.")
@@ -45,31 +47,79 @@ def cleanup_input(words: str) -> set[str]:
 
 
 def contains_punctuation(word):
-    """Check to see if puncuation is present in the provided string."""
+    """Check to see if punctuation is present in the provided string."""
     return any([True if c in string.punctuation else False for c in word])
 
 
-def highlight_solution(
-    puzzle: Puzzle,
-    solution: Optional[Puzzle] = None,
-) -> Puzzle:
-    """Convert puzzle array of nested lists into a string.
+def is_palindrome(word: str) -> bool:
+    """Check is a word in a palindrome."""
+    return word == word[::-1]
 
-    Args:
-        puzzle (Puzzle): he current puzzle state.
-        solution (Optional[Puzzle], optional): If a solution is provided
-        it will be highlighted in the output. Defaults to None.
 
-    Returns:
-        str: The current puzzle as a string.
-    """
-    output = []
-    for r, line in enumerate(puzzle):
+def word_contains_word(words: Wordlist, word: str) -> bool:
+    """Make sure `test_word` cannot be found in any word
+    in `words`, going forward or backward."""
+    for test_word in words:
+        if (
+            word in test_word.text.upper()
+            or word[::-1] in test_word.text.upper()
+            or test_word.text.upper() in word
+            or test_word.text.upper()[::-1] in word
+        ):
+            return True
+    return False
+
+
+def validate_direction_iterable(
+    d: Iterable[str | Tuple[int, int] | Direction]
+) -> DirectionSet:
+    """Validates that all the directions in d are found as keys to
+    config.dir_moves and therefore are valid directions."""
+    o = set()
+    for direction in d:
+        if isinstance(direction, Direction):
+            o.add(direction)
+            continue
+        elif isinstance(direction, tuple):
+            o.add(Direction(direction))
+            continue
+        try:
+            o.add(Direction[direction.upper().strip()])
+        except KeyError:
+            raise ValueError(f"'{direction}' is not a valid direction.")
+    return o
+
+
+def validate_level(d) -> DirectionSet:
+    """Given a d, try to turn it into a list of valid moves."""
+    if isinstance(d, int):  # traditional numeric level
+        try:
+            return config.level_dirs[d]
+        except KeyError:
+            raise ValueError(
+                f"{d} is not a valid difficulty number"
+                + f"[{', '.join([str(i) for i in config.level_dirs.keys()])}]"
+            )
+    if isinstance(d, str):  # comma-delimited list
+        return validate_direction_iterable(d.split(","))
+    if isinstance(d, Iterable):  # probably used by external code
+        return validate_direction_iterable(d)
+    raise TypeError(f"{type(d)} given, not str, int, or Iterable[str]\n{d}")
+
+
+def direction_set_repr(ds: DirectionSet) -> str:
+    return ("'" + ",".join(d.name for d in ds) + "'") if ds else "None"
+
+
+def highlight_solution(puzzle: WordSearch) -> Puzzle:
+    """Add highlighting to puzzle solution."""
+    output: Puzzle = []
+    for r, line in enumerate(puzzle.puzzle):
         line_chars = []
-        # check to see if character if part of a puzzle word
+        # check to see if character if part of the solution
         for c, char in enumerate(line):
-            if solution and solution[r][c] == "•":
-                line_chars.append(f"{Style.DIM}{char}{Style.RESET_ALL}")
+            if puzzle.solution[r][c]:
+                line_chars.append(f"\u001b[1m\u001b[31m{char}\u001b[0m")
             else:
                 line_chars.append(f"{char}")
         output.append(line_chars)
@@ -77,16 +127,8 @@ def highlight_solution(
 
 
 def make_header(puzzle: Puzzle, text: str) -> str:
-    """Generate a header that fits the current puzzle.
-
-    Args:
-        puzzle (Puzzle): The current puzzle state.
-        text (str): The text to include in the header.
-
-    Returns:
-        str: Formatted header.
-    """
-    hr = "-" * (len(puzzle) * 2 - 1)
+    """Generate a header that fits the current puzzle."""
+    hr = "-" * max(11, (len(puzzle) * 2 - 1))
     padding = " " * ((len(hr) - len(text)) // 2)
     return f"""{hr}
 {padding}{text}{padding}
@@ -94,45 +136,55 @@ def make_header(puzzle: Puzzle, text: str) -> str:
 
 
 def stringify(puzzle: Puzzle) -> str:
-    """Convert puzzle array of nested lists into a string.
-
-    Args:
-        puzzle (Puzzle): The current puzzle state.
-
-    Returns:
-        str: The current puzzle as a string.
-    """
+    """Convert puzzle array of nested lists into a string."""
     output = []
     for line in puzzle:
         output.append(" ".join(line))
     return "\n".join(output)
 
 
-def replace_right(
-    string: str, target: str, replacement: str, replacements: int = 1
-) -> str:
-    """Replace `target` with `replacement` from the right size of the string."""
-    return replacement.join(string.rsplit(target, replacements))
+def format_puzzle_for_show(puzzle: WordSearch, show_solution: bool = False) -> str:
+    header = make_header(puzzle.puzzle, "WORD SEARCH")
+    word_list = get_word_list_str(puzzle.key)
+    # highlight solution if provided
+    puzzle_list = highlight_solution(puzzle) if show_solution else puzzle.puzzle
+    return f"""{header}
+{stringify(puzzle_list)}
+
+Find these words: {word_list if word_list else '<ALL SECRET WORDS>'}
+* Words can go {get_level_dirs_str(puzzle.level)}.
+
+Answer Key: {get_answer_key_str(puzzle.placed_words)}"""
 
 
-def get_level_dirs_str(level: int) -> str:
-    """Return all pozzible directions for specified level."""
-    return replace_right(", ".join(config.level_dirs[level]), ", ", ", and ")
+def get_level_dirs_str(level: DirectionSet) -> str:
+    """Return possible directions for specified level as a string."""
+    level_dirs_str = [d.name for d in level]
+    level_dirs_str.insert(-1, "and")
+    return ", ".join(level_dirs_str)
 
 
 def get_word_list_str(key: Key) -> str:
-    """Return all placed puzzle words as a list."""
-    return ", ".join([k for k in sorted(key.keys())])
+    """Return all placed puzzle words as a list (excluding secret words)."""
+    return ", ".join(get_word_list_list(key))
 
 
-def get_answer_key_str(key: Key) -> str:
+def get_word_list_list(key: Key) -> List[str]:
+    """Return all placed puzzle words as a list (excluding secret words)."""
+    return [k for k in sorted(key.keys()) if not key[k]["secret"]]
+
+
+def get_answer_key_list(words: Wordlist) -> List[Any]:
     """Return a easy to read answer key for display/export."""
     keys = []
-    for k in sorted(key.keys()):
-        direction = key[k]["direction"]
-        raw_coords: tuple[int, int] = key[k]["start"]
-        coords = (raw_coords[0] + 1, raw_coords[1] + 1)
-        keys.append(f"{k} {direction} @ {coords}")
+    for w in sorted(words, key=lambda word: word.text):
+        keys.append(w.key_string)
+    return keys
+
+
+def get_answer_key_str(words: Wordlist) -> str:
+    """Return a easy to read answer key for display."""
+    keys = get_answer_key_list(words)
     return ", ".join(keys)
 
 
@@ -421,7 +473,6 @@ WORD_LIST = [
     "experience",
     "expert",
     "explain",
-    "eye",
     "face",
     "fact",
     "factor",
@@ -594,7 +645,6 @@ WORD_LIST = [
     "less",
     "let",
     "letter",
-    "level",
     "lie",
     "life",
     "light",
